/* dnsmasq is Copyright (c) 2000 Simon Kelley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 dated June, 1991.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

/* See RFC1035 for details of the protocol this code talks. */

/* Author's email: simon@thekelleys.org.uk */

#include <arpa/nameser.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <syslog.h>
#include <getopt.h>

#define VERSION "0.5"

struct crec { 
  char name[MAXDNAME];
  unsigned char addr[INADDRSZ];
  struct crec *next, *prev;
};

static void cache_insert(struct crec *crecp);
static struct crec *cache_get_free(void);
static void cache_unlink (struct crec *crecp);
static int process_request(HEADER *header, unsigned char *limit);
static void do_reverse_lookup(HEADER *header, char *name, unsigned char *namep,
			      unsigned char *limit, unsigned char **anspp);
static void do_forward_lookup(HEADER *header, char *name, unsigned char *namep,
			      unsigned char *limit, unsigned char **anspp);
static void do_mx_lookup(HEADER *header, char *name, unsigned char *namep,
			 unsigned char *limit, unsigned char **anspp);
static void multiplex(void);

static char *mxname;
static int port;
static struct crec *cache_head, *cache_tail;

int main (int argc, char **argv)
{
  char option;
  int i, cachesize;
  struct crec *crecp;

  mxname = NULL;
  port = NAMESERVER_PORT;
  cachesize = 200;

  opterr = 0;

  while (1)
    {
      option = getopt(argc, argv, "vm:p:c:");
      
      if (option == 'v')
	{
	  fprintf(stderr, "dnsmasq version %s\n", VERSION);
	  exit(0);
	}

      if (option == 'm')
	{
	  mxname = malloc(strlen(optarg)+1);
	  strcpy(mxname, optarg);
	}

      if (option == 'c')
	{
	  cachesize = atoi(optarg);
	  if (cachesize < 20)
	    cachesize = 20;

	  if (cachesize > 1000)
	    cachesize = 1000;
	}

      if (option == 'p')
	port = atoi(optarg);
      
      if (option == '?')
	{ 
	  fprintf(stderr, 
		  "Usage: dnsmasq -v -p <port> -m <mxhost> -c <cachesize>\n");
	  exit(0);
	}
      
      if (option == -1)
	break;
    }

  crecp = (struct crec *)malloc(cachesize*sizeof(struct crec));
  cache_head = NULL;
  cache_tail = crecp;
  for (i=0; i<cachesize; i++, crecp++)
    {
      cache_insert(crecp);
      crecp->name[0] = 0; /* empty entry */
    }

  /* The following code "daemonizes" the process. See Stevens section 12.4 */
 
  if (fork() != 0 )
    exit(0);

  setsid();
  signal(SIGHUP, SIG_IGN);

  if (fork() != 0)
    exit(0);

  chdir("/");
  umask(0);

  for (i=0; i<64; i++)
    close(i);
  
  openlog("dnsmasq", LOG_PID, LOG_DAEMON);

  syslog(LOG_INFO, "started, version %s cachesize %d", VERSION, cachesize);
  if (mxname)
    syslog(LOG_INFO, "serving MX record for mailhost %s", mxname);
  
  multiplex();

  return 0;
}

static void multiplex()
{
  int udpfd, listenfd, confd;
  struct sockaddr_in nsaddr, udpaddr, tcpaddr;
  fd_set rset, wset;
  socklen_t udplen, tcplen;
  /* Size: we check after adding each record, so there must be 
     memory for the largest packet, and the largest record */
  char packet[PACKETSZ+MAXDNAME+RRFIXEDSZ];
  u_int16_t reqsize, repsize;
  int repmax;
  char *tcppacket, *readp, *writep;
  enum { ACCEPT, READL0, READL1, READ, WRITEL0, WRITEL1, WRITE } state;

  udpfd = socket(AF_INET, SOCK_DGRAM, 0);
  listenfd = socket(AF_INET, SOCK_STREAM, 0);
  if ((udpfd == -1) || (listenfd == -1))
    {
      syslog(LOG_ERR,"cannot create socket: %m");
      exit(1);
    }

  bzero(&nsaddr, sizeof(nsaddr));
  nsaddr.sin_family = AF_INET;
  nsaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  nsaddr.sin_port = htons(port);

  if (bind(udpfd, (struct sockaddr *) &nsaddr, sizeof(nsaddr)) ||
      bind(listenfd, (struct sockaddr *) &nsaddr, sizeof(nsaddr)) ||
      listen(listenfd, 5))
    {
      syslog(LOG_ERR, "bind failed: %m");
      exit(1);
    }

  state = ACCEPT;
  while (1)
    {
      int maxfd = udpfd;
      
      FD_ZERO(&rset);
      FD_ZERO(&wset);
      FD_SET(udpfd, &rset);
      
      switch (state)
	{
	case ACCEPT:
      	  FD_SET(listenfd, &rset);
	  if (listenfd > maxfd)
	    maxfd = listenfd;
	  break;
	case READL0:
	case READL1:
	case READ:
	  FD_SET(confd, &rset);
	  if (confd > maxfd)
	    maxfd = confd;
	  break;
	case WRITEL0:
	case WRITEL1:
	case WRITE:
	  FD_SET(confd, &wset);
	  if (confd > maxfd)
	    maxfd = confd;
	  break;
	}
      
      if (select(maxfd+1, &rset, &wset, NULL, NULL) == -1)
	continue;

      if (FD_ISSET(udpfd, &rset))
	{
	  int n = recvfrom(udpfd, packet, PACKETSZ, 0, &udpaddr, &udplen); 
	  if (n > sizeof(HEADER))
	    {
	      n = process_request((HEADER *)packet, packet + PACKETSZ);
	      sendto(udpfd, packet, n, 0, &udpaddr, udplen);  
	    }
	}
      
      switch (state)
	{
	case ACCEPT:
	  if (FD_ISSET(listenfd, &rset))
	    {
	      confd = accept(listenfd, &tcpaddr, &tcplen);
	      if (confd != -1)
		state = READL0;
	    }
	  break;

	case READL0:
	  if (FD_ISSET(confd, &rset))
	    {
	      unsigned char c;
	      if (read(confd, &c, 1) == 1)
		{
		  reqsize = c<<8;
		  state = READL1;
		}
	    }
	  break;

	case READL1:
	  if (FD_ISSET(confd, &rset))
	    {
	      int val;
	      unsigned char c;
	      if (read(confd, &c, 1) == 1)
		{
		  reqsize |= c;
		  /* Size: we check after adding each record, so there must
		     be memory for the largest record in excess of the
		     limit. We arbritrarily limit the reply size to
		     four times the query size. */
		  repmax = reqsize*4;
		  tcppacket = readp = malloc(repmax+MAXDNAME+RRFIXEDSZ);
		  if (!tcppacket)
		    {
		      close(confd);
		      state = ACCEPT;
		    }
		  else
		    {
		      val = fcntl(confd, F_GETFL, 0); /* set non-blocking */
		      fcntl(confd, F_SETFL, val | O_NONBLOCK);
		      state = READ;
		    }
		}
	    }
	  break;

	case READ:
	  if (FD_ISSET(confd, &rset))
	    {
	      int n = read(confd, readp, reqsize);
	      if (n > 0)
		{ 
		  readp += n;
		  reqsize -= n;
		  if (reqsize == 0)
		    {
		      repsize = process_request((HEADER *)tcppacket,
						tcppacket + repmax);
		      writep = tcppacket;
		      state = WRITEL0;
		    }
		}
	    }
	  break;
	
	case WRITEL0:
	  if (FD_ISSET(confd, &wset))
	    {
	      unsigned char c = repsize >> 8;
	      if (write(confd, &c, 1) == 1)
		state = WRITEL1;
	    }
	  break;
	
	case WRITEL1:
	  if (FD_ISSET(confd, &wset))
	    {
	      unsigned char c = repsize & 255;
	      if (write(confd, &c, 1) == 1)
		state = WRITE;
	    }
	  break;
	
	case WRITE:
	  if (FD_ISSET(confd, &wset))
	    {
	      int n = write(confd, writep, repsize);
	      if (n > 0)
		{
		  writep += n;
		  repsize -= n;
		  if (repsize == 0)
		    {
		      free(tcppacket);
		      close(confd);
		      state = ACCEPT;
		    }
		}
	    }
	  break;
      	}
    }
}

static int process_request(HEADER *header, unsigned char *limit)
{
  unsigned char *ansp, *namep, *p, *p1 ;
  int qtype, qclass;
  char name[MAXDNAME];
  char *cp;
  unsigned int j,l,q;
  int qdcount = ntohs(header->qdcount);
  int opcode = header->opcode;

  header->ancount = htons(0);
  header->nscount = htons(0);
  header->arcount = htons(0);

  header->tc = 0;

  if (!qdcount)
    {
      header->rcode = htons(FORMERR);
      header->qr = 1;
      header->ra = 1;
      return sizeof(HEADER);
    }
  
  /* determine end of question section (we put answers there) */
  p = ansp = (char *)(header+1);
  for (q=0; q<qdcount; q++)
    {
      while (1)
	{
	  if (((*ansp) & 0xc0) == 0xc0) /* pointer for name compression */
	    {
	      ansp += 2;
	      break;
	    }
	  else if (*ansp) /* another segment */
	    ansp += (*ansp) + 1;
	  else            /* end */
	    {
	      ansp++;
	      break;
	    }
	}
      ansp += 4; /* class and type */
    }

  /* now process each question, answers go in RRs after the question */
  for (q=0; q<qdcount; q++)
    {
      /* save pointer to name for copying into answers */
      namep = p;

      /* now extract name as .-concatenated string into name */
      cp = name;
      p1 = NULL;
      while ((l = *p++))
	{
	  if ((l & 0xc0) == 0xc0) /* pointer */
	    { 
	      /* get offset */
	      l = (l&0x3f) << 8;
	      l |= *p++;
	      if (!p1) /* first jump, save location to go back to */
		p1 = p;
	      p = l + (unsigned char *)header;
	    }
	  else
	    {
	      for(j=0; j<l; j++)
		*cp++ = tolower(*p++);
	      *cp++ = '.';
	    }
	}
      *--cp = 0; /* overwrite last period */
	  
      if (p1) /* we jumped via compression */
	p = p1;

      GETSHORT(qtype, p); 
      GETSHORT(qclass, p);
  
      if ((opcode == QUERY) && (qclass == C_IN) && (qtype == T_PTR))
	do_reverse_lookup(header, name, namep, limit, &ansp);
      else if ((opcode == QUERY) && (qclass == C_IN) && (qtype == T_A))
	do_forward_lookup(header, name, namep, limit, &ansp);
      else if ((opcode == QUERY) && (qclass == C_IN) && (qtype == T_MX))
	do_mx_lookup(header, name, namep, limit, &ansp);
      
    }
  /* done all questions, set up header and return result */
  
  header->qr = 1; /* response */
  header->aa = 0; /* authoritive - never */
  header->ra = 1; /* recursion if available */
  header->rcode = htons(NOERROR); /* no error */
  return ansp - (unsigned char *)header;
}


static void do_forward_lookup(HEADER *header, char *name, unsigned char *namep,
			      unsigned char *limit, unsigned char **anspp)
{
  unsigned char *p = *anspp;
  struct crec *crecp = cache_head;
  struct hostent *ent;
  int i;
  
  /* first search the cache for an answer */
  while (crecp)
    { 
      if  (strcmp(crecp->name, name) == 0)
	{
	  do 
	    {
	      struct crec *tmp = crecp->next;
	      /* promote this back up to the youngest end of the
		 cache list */
	      cache_unlink(crecp);
	      cache_insert(crecp);
	      crecp = tmp;
	    } 
	  /* keep going for multiple entries with same name */
	  while (crecp && (strcmp(crecp->name, name) == 0));
	  goto incache;
	}
      crecp = crecp->next;
    }

  /* not found in cache: go to server */
  
  ent = gethostbyname(name);

  if (!ent)
    return; /* failed */

  /* insert new cache entries at the head of the list */
  for(i=0; ent->h_addr_list[i]; i++)
    {
      crecp = cache_get_free();
      strcpy(crecp->name, name);
      memcpy(&crecp->addr, ent->h_addr_list[i], ent->h_length);
      cache_insert(crecp);
    }

incache:
  /* the entrie(s) we want are guaranteed to be at the head of the list */
  for(crecp = cache_head; strcmp(name, crecp->name) == 0; crecp = crecp->next)
    {
      /* copy question as first part of answer (use compression)*/
  
      PUTSHORT((namep - (unsigned char *)header) | 0xc000, p); 
      PUTSHORT(T_A, p);
      PUTSHORT(C_IN, p);
      PUTLONG(0, p); /* TTL -- zero means this transaction only */
      
      PUTSHORT(INADDRSZ, p);
      memcpy(p, &crecp->addr, INADDRSZ);
      p += INADDRSZ;
     
      if ((limit - p) < 0)
	{
	  /* last answer exceeded packet size, set truncated bit */
	  header->tc = 1;
	  break;
	}
      else
	{
	  header->ancount = htons(ntohs(header->ancount)+1);
	  *anspp = p;
	}
    }
}


static void do_reverse_lookup(HEADER *header, char *name, unsigned char *namep,
			      unsigned char *limit, unsigned char **anspp)
{
  struct hostent *ent;
  unsigned char *p = *anspp;
  unsigned char *sav;
  char *cp1, *cp2;
  int i,j;
  unsigned char addr[4];
  struct crec *crecp = cache_head;

  /* turn name into a series of asciiz strings */
  /* j counts no of labels */
  for(j = 1,cp1 = name; *cp1; cp1++)
    if (*cp1 == '.')
      {
	*cp1 = 0;
	j++;
      }
    
  addr[0] = addr[1] = addr[2] = addr[3] = 0;

  /* address arives as a name of the form
     www.xxx.yyy.zzz.in-addr.arpa
     some of the low order address octets might be missing
     and should be set to zero. */
  for (cp1 = name,i=0; i<j; i++)
    {
      if (strcmp(cp1, "in-addr") == 0)
	break;
      addr[3] = addr[2];
      addr[2] = addr[1];
      addr[1] = addr[0];
      addr[0] = atoi(cp1);
      cp1 += strlen(cp1)+1;
    }

  /* first search the cache for an answer */
  while (crecp)
    { 
      if  (*(crecp->name) && (memcmp(crecp->addr, addr, INADDRSZ) == 0))
	{
	  /* preserve cache invariant that say all aliases of a name
	     are adjecent to each other by promoting all aliases to
	     head of cache. Note that the exact reverse mapping we want
	     might not end up precisely as the head entry, but we 
	     still have a pointer to it in crecp. */
	  struct crec *tmp = crecp->prev;
	  while (tmp && (strcmp(crecp->name, tmp->name) == 0))
	    {
	      struct crec *tmp1 = tmp->prev;
	      cache_unlink(tmp);
	      cache_insert(tmp);
	      tmp = tmp1;
	    }
	  tmp = crecp;
	  while (tmp && (strcmp(crecp->name, tmp->name) == 0))
	    {
	      struct crec *tmp1 = tmp->next;
	      cache_unlink(tmp);
	      cache_insert(tmp);
	      tmp = tmp1;
	    }
	  goto incache;
	}
      crecp = crecp->next;
    }

  /* not in cache: hit server */
  ent = gethostbyaddr(addr, 4, AF_INET);
  
  if (!ent) /* failed lookup */
    return;
  
  /* insert new cache entry at head of list */
  crecp = cache_get_free();
  strcpy(crecp->name, ent->h_name);
  memcpy(&crecp->addr, addr, INADDRSZ);
  cache_insert(crecp);

incache:
  /* copy question as first part of answer (use compression) */
  
  PUTSHORT((namep - (unsigned char *)header) | 0xc000, p); 
  PUTSHORT(T_PTR, p);
  PUTSHORT(C_IN, p);
  PUTLONG(0, p); /* TTL -- zero means this transaction only */
  
  sav = p;
  PUTSHORT(0, p); /* dummy RDLENGTH */
  cp1 = crecp->name;
  while (*cp1) 
    {
      cp2 = p++;
      for (j=0; *cp1 && (*cp1 != '.'); cp1++, j++)
	*p++ = *cp1;
      *cp2 = j;
      if (*cp1)
	cp1++;
    }
  *p++ = 0;
  j = p - sav - 2;
  PUTSHORT(j, sav); /* Real RDLENGTH */
   
  if ((limit - p) < 0)
    /* last answer exceeded packet size, set truncated bit */
    header->tc = 1;
  else
    {
      header->ancount = htons(ntohs(header->ancount)+1);
      *anspp = p;
    }
}

static void do_mx_lookup(HEADER *header, char *name, unsigned char *namep,
			 unsigned char *limit, unsigned char **anspp)
{
  unsigned char *p = *anspp;
  unsigned char *sav, *hostnamep;
  char hostname[MAXDNAME];
  char *cp1, *cp2;
  int j;
  
  if (!mxname)
    return;

  if (gethostname(hostname, MAXDNAME) != 0)
    return;

  /* we return an mx record for a name given on the command line */
  if (strcmp(name, mxname) != 0)
    return;
  
  PUTSHORT((namep - (unsigned char *)header) | 0xc000, p);       
  PUTSHORT(T_MX, p);
  PUTSHORT(C_IN, p);
  PUTLONG(0, p); /* TTL -- zero means this transaction only */
      
  sav = p;
  PUTSHORT(0, p); /* dummy RDLENGTH */
  
  PUTSHORT (1, p); /* preference */

  cp1 = hostname;
  hostnamep = p;
  while (*cp1) 
    {
      cp2 = p++;
      for (j=0; *cp1 && (*cp1 != '.'); cp1++, j++)
	*p++ = *cp1;
      *cp2 = j;
      if(*cp1)
	cp1++;
    }
  *p++ = 0;
  j = p - sav - 2;
  PUTSHORT(j, sav); /* Real RDLENGTH */

  if ((limit - p) < 0)
    /* last answer exceeded packet size, set truncated bit */
    header->tc = 1;
  else
    {
      header->ancount = htons(ntohs(header->ancount)+1);
      *anspp = p;
  
      /* add an A answer for the host we just returned as answer to MX query */
      do_forward_lookup(header, hostname, hostnamep, limit, anspp);
    } 
}

/* insert a new cache entry at the head of the list (youngest entry) */
static void cache_insert(struct crec *crecp)
{
  if (cache_head)
    cache_head->prev = crecp;
  crecp->next = cache_head;
  crecp->prev = NULL;
  cache_head = crecp;
}

/* get a cache entry to re-cycle from the tail of the list (oldest entry) */
static struct crec *cache_get_free(void)
{
  struct crec *ret = cache_tail;
  struct crec *crecp;

  cache_tail->prev->next = NULL;
  crecp = cache_tail->prev;
  cache_tail = cache_tail->prev;

  /* the next bit is non-obvious, it ensures that if the cache
     hold more than one address for a name, they are all dumped when 
     the first entry is re-used. If this is not done, a subset of the
     addresses could be returned from the cache, rather than a new
     lookup being triggered */

  while (crecp->name[0] && (strcmp(ret->name, crecp->name) == 0))
    {
      crecp->name[0] = 0;
      crecp = crecp->prev;
    }

  return ret;
}

/* remove an arbitrary cache entry for promotion */ 
static void cache_unlink (struct crec *crecp)
{
  if (crecp->prev)
    crecp->prev->next = crecp->next;
  else
    cache_head = crecp->next;

  if (crecp->next)
    crecp->next->prev = crecp->prev;
  else
    cache_tail = crecp->prev;
}








